local L = Capitalist.Locale
local realm = GetRealmName()
local character = UnitName("player")

local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")

-- products are not allowed to have costs set directly
-- update data in case products were discovered after
-- costs were previously set on the product
-- (i.e. it was treated as a resource)
function Capitalist:EnforceProductsCannotHaveCostSetting()
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			for skillName, skillData in pairs(professionData) do
				local productName = self:ProductNameFromSkillName(profession, skillName)
				if (not self:SkillIsCyclic(profession, skillName) and
					self:IsComponent(productName)) then

					local cost = self.db.profile.components[productName].cost
					if (cost) then
						local costCopper = cost * 10000	-- TODO: 10000
						local costString = self:PrettyCurrency(costCopper)
						self:Print(string.format("  "..L["Removing cost (%s) from product (%s)."], costString, self:GetComponentText(productName)))
						self.db.profile.components[productName].cost = nil
					end

					local buyPrice = self.db.profile.components[productName].buyPrice
					if (buyPrice) then
						local buyPriceString = self:PrettyCurrency(buyPrice)
						self:Print(string.format("  "..L["Removing buy price (%s) from product (%s)."], buyPriceString, self:GetComponentText(productName)))
						self.db.profile.components[productName].buyPrice = nil
					end
				end
			end
		end
	end
end

function Capitalist:ApplyFilters()
	self:DebugEntry("ApplyFilters")

	-- reset memory
	self:ResourceMemoryReset()
	self:PartMemoryReset()
	self:ProductMemoryReset()

	-- CONSIDER:  Automatic queueing of parts that alts need to make

	local productsToProcess = {}
	local resourcesToMakeVisible = {}

	local q = 0

	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			productsToProcess[profession] = { skills = {}, pending = 1, }

			self:DebugMisc(profession)

			for skillName, skillData in pairs(professionData) do
				-- for products not allowed, mark them as blocked
				local productBlocked
				if (not self:ProductIsAllowed(profession, skillName)) then
					productBlocked = true
				end

				-- for products with unavailable product name, block as well
				local productName = self:ProductNameFromSkillName(profession, skillName)
				if (not productName) then
					productBlocked = true
				end

				self:ProductSetVisible(profession, skillName, not productBlocked)
				if (not productBlocked) then
					if (self:ProductIsEnabled(profession, skillName)) then
						table.insert(productsToProcess[profession].skills, skillName)
					end
				end
			end
		end
	end

	local doneWithProducts
	while (not doneWithProducts) do
		doneWithProducts = true
		for profession, professionData in pairs(self.db.profile.professions) do
			if (productsToProcess[profession].skills) then
				while (productsToProcess[profession].pending <= #productsToProcess[profession].skills) do
					local skillName = productsToProcess[profession].skills[productsToProcess[profession].pending]
					local product = self.db.profile.professions[profession][skillName]
					for componentName, componentCount in pairs(product.components) do
						local componentSkillName, componentProfession = self:SkillNameFromProductName(componentName)
						if (componentSkillName and componentProfession) then
							self:PartSetVisible(componentName, true)
							if (not tContains(productsToProcess[componentProfession].skills, componentSkillName)) then
								self:ProductSetVisible(componentProfession, componentSkillName, true)
								table.insert(productsToProcess[componentProfession].skills, componentSkillName)
								doneWithProducts = false
							end
						else
							if (not tContains(resourcesToMakeVisible, componentName)) then
								table.insert(resourcesToMakeVisible, componentName)
							end
						end
					end

					-- enchanting-specific
					local productName = self:ProductNameFromSkillName(profession, skillName)
					if (productName:find(L["Scroll of Enchant "])) then
						local componentName = L["Enchanting Vellum"]
						local componentSkillName, componentProfession = self:SkillNameFromProductName(componentName)
						if (componentSkillName and (not tContains(productsToProcess[componentProfession].skills, componentSkillName))) then
							self:ProductSetVisible(componentProfession, componentSkillName, true)
							table.insert(productsToProcess[componentProfession].skills, componentSkillName)
							doneWithProducts = false
						end
					end

					productsToProcess[profession].pending = productsToProcess[profession].pending + 1
				end
			end
		end
	end

	for i=1, #resourcesToMakeVisible do
		-- make visible resources that are in use
		local resourceName = resourcesToMakeVisible[i]
		if (self:ResourceGetMerchant(resourceName)) then
			-- resources with merchants suppliers don't have
			-- a separate cost value; they use the buyPrice
			if (self:HasResourceCopperCost(resourceName)) then
				local cost = self:GetResourceCopperCost(resourceName)
				local costString = self:PrettyCurrency(cost)
				self:Print(string.format("  "..L["Removing cost (%s) from merchant-supplied resource (%s)."], costString, self:GetComponentText(resourceName)))
				self:SetResourceCopperCost(resourceName, nil)
			end
		else
			self:ResourceSetVisible(resourceName, true)
		end
	end

	self:BuildList()
	self:UpdateProductOptions()
	self:UpdateResourceCosts()
	self:UpdateStatistics()

	self:DebugExit("ApplyFilters")
end

-- scans the trade skill frame and adds all products the player knows
function Capitalist:ScanTradeSkillFrame(...)
	self:DebugEntry("ScanTradeSkillFrame")

	local profession = ...;
	self:DebugMisc(profession)

	if (profession == L["Runeforging"]) then
		return self:DebugExit("ScanTradeSkillFrame")
	end

	-- REVIEW: Update products for other players, and maybe track their names to do business later
	-- Bypass data updating when viewing another player's profession information
	if (IsTradeSkillLinked()) then
		return self:DebugExit("ScanTradeSkillFrame")
	end

	SetTradeSkillSubClassFilter(0, 1, 1)
	SetTradeSkillInvSlotFilter(0, 1, 1)
	SetTradeSkillItemNameFilter("")

	-- REVIEW: Workaround for cache not full ready yet
	local _, skillTypeHeader, _, _, _ = GetTradeSkillInfo(1)
	if (skillTypeHeader ~= "header") then
		return self:DebugExit("ScanTradeSkillFrame")
	end
	
	if (not self.db.profile[realm]) then
		self.db.profile[realm] = {}
	end
	if (not self.db.profile[realm][character]) then
		self.db.profile[realm][character] = {}
	end
	if (not self.db.profile[realm][character][profession]) then
		self.db.profile[realm][character][profession] = {}
	end

	local newProducts = 0
	local newComponents = 0
	local newCharacterProducts = 0

	if (not self.db.profile.professions) then
		self.db.profile.professions = {}
	end
	if (not self.db.profile.professions[profession]) then
		self.db.profile.professions[profession] = {}
	end
	
	if (not self.db.profile.components) then
		self.db.profile.components = {}
	end
	
	for skillId=1, GetNumTradeSkills() do

		local skillName, skillType, numAvailable, isExpanded, altVerb = GetTradeSkillInfo(skillId)
		self:DebugMisc(skillName)		-- "Enchant Bracers - Major Stamina"
		self:DebugMisc(skillType)		-- "header" or one of ("trivial", "easy", "medium", "optimal", "difficult")
		self:DebugMisc(numAvailable)	-- the bracketed value shown in the craft UI, showing how many you can make with current components
		self:DebugMisc(isExpanded)		--
		self:DebugMisc(altVerb)			-- nil or "Enchant" (REVIEW: perhaps others for armor kits?)

		local itemLink = GetTradeSkillItemLink(skillId)
		self:DebugMisc(itemLink)

		local recipeLink = GetTradeSkillRecipeLink(skillId)
		self:DebugMisc(recipeLink)

		-- REVIEW: Correction for two glyphs named "Glyph of Death Coil"
		if (skillName == "Glyph of Death Coil") then
			local recipeIdString = string.match(recipeLink, "Henchant:(%d+)")
			local recipeIdNumber = tonumber(recipeIdString)
			-- item 45804, spell 64266 - DK
			if (recipeIdNumber == 64266) then
				skillName = skillName.." <Death Knight>"
			end
			-- item 42457, spell 57261 - Warlock
			if (recipeIdNumber == 57261) then
				skillName = skillName.." <Warlock>"
			end
		end

		if ((skillType ~= "header") and L.checkForValidSkillName(profession, skillName)) then
			local skillDescription = GetTradeSkillDescription(skillId)
			self:DebugMisc(skillDescription)

			local minMade, maxMade = GetTradeSkillNumMade(skillId)
			self:DebugMisc(minMade)
			self:DebugMisc(maxMade)

			if (maxMade == minMade) then
				maxMade = nil
				if (minMade == 1) then
					minMade = nil
				end
			end

			-- add it to the known products
			if (not self.db.profile.professions[profession][skillName]) then
				self.db.profile.professions[profession][skillName] = {
																		components = {},
																	 }

				newProducts = newProducts + 1
			end

			-- add the known products for this character
			if (not self.db.profile[realm][character][profession][skillName]) then
				self.db.profile[realm][character][profession][skillName] = { }
				newCharacterProducts = newCharacterProducts + 1
			end


			-- CONSIDER: For all information extractable from link, itemLink, or recipeLink *don't store it*.

			if (self.db.profile.professions[profession][skillName].description ~= skillDescription) then
				self.db.profile.professions[profession][skillName].description = skillDescription
			end

			if (self.db.profile.professions[profession][skillName].itemLink ~= itemLink) then
				self.db.profile.professions[profession][skillName].itemLink = itemLink
			end

			if (self.db.profile.professions[profession][skillName].recipeLink ~= recipeLink) then
				self.db.profile.professions[profession][skillName].recipeLink = recipeLink
			end

			if (self.db.profile.professions[profession][skillName].min ~= minMade) then
				self.db.profile.professions[profession][skillName].min = minMade
			end
			if (self.db.profile.professions[profession][skillName].max ~= maxMade) then
				self.db.profile.professions[profession][skillName].max = maxMade
			end

			if (self.db.profile[realm][character][profession][skillName].type ~= skillType) then
				self.db.profile[realm][character][profession][skillName].type = skillType
			end

			if (itemLink) then
				local _, _, _, _, _, _, _, _, _, _, itemSellPrice = GetItemInfo(itemLink)
				if (itemSellPrice and (itemSellPrice > 0)) then
					if (self.db.profile.professions[profession][skillName].sellPrice ~= itemSellPrice) then
						self.db.profile.professions[profession][skillName].sellPrice = itemSellPrice
					end
				end
			end

			local componentsScanned = {}
			local componentsComplete = false

			for reagentId=1, GetTradeSkillNumReagents(skillId) do
				local reagentName, reagentTexture, reagentCount, playerReagentCount = GetTradeSkillReagentInfo(skillId, reagentId)
				self:DebugMisc(reagentName)
				self:DebugMisc(reagentTexture)
				self:DebugMisc(reagentCount)
				self:DebugMisc(playerReagentCount)

				if (reagentName and (reagentCount > 0)) then
					componentsComplete = true
					componentsScanned[reagentName] = reagentCount

					-- for previously unknown components, save the item link and the id
					if (not self.db.profile.components[reagentName]) then

						local reagentLink = GetTradeSkillReagentItemLink(skillId, reagentId)
						self:DebugMisc(reagentLink)

						-- TODO: Optimization - don't store this ID
						local _, _, reagentItemId = reagentLink:find("Hitem:(%d+):")
						self:DebugMisc(reagentItemId)
					
						self.db.profile.components[reagentName] = {
																	link = reagentLink,
																	id = tonumber(reagentItemId),
																  }

						newComponents = newComponents + 1
					end
				else
					componentsComplete = false
					break
				end
			end

			if (componentsComplete) then
				local dbComponents = self.db.profile.professions[profession][skillName].components

				for oldComponentName, oldComponentCount in pairs(dbComponents) do
					local newComponentCount = componentsScanned[oldComponentName]
					if (not newComponentCount) then
						dbComponents[oldComponentName] = nil
					elseif (newComponentCount ~= oldComponentCount) then
						dbComponents[oldComponentName] = componentsScanned[oldComponentName]
						componentsScanned[oldComponentName] = nil
					else
						componentsScanned[oldComponentName] = nil
					end
				end

				for newComponentName, newComponentCount in pairs(componentsScanned) do
					dbComponents[newComponentName] = newComponentCount
				end
			end
		end
	end

	-- REVIEW: Still needed?  If so, make smarter about only apply to enchants.
	if (profession == L["Enchanting"]) then
		local componentName = L["Enchanting Vellum"]
		if (not self.db.profile.components[componentName]) then
			local itemLink = select(2, GetItemInfo(self:GetComponentId("Enchanting Vellum")))
			self.db.profile.components[componentName] = {
															id = self:GetComponentId(componentName),
															link = itemLink,
														}
			newComponents = newComponents + 1
		end
	end

	if (newProducts + newComponents + newCharacterProducts > 0) then
		self:Print(string.format(
						L["Found %d new products and %d new components."],
						newProducts,
						newComponents
					))
		self:Print(string.format(
						L["Found %d new character products."],
						newCharacterProducts
					))

		self:EnforceProductsCannotHaveCostSetting()
		self:UpdateComponentValues()
		self:UpdateCombinedValues()
	
		--TODO: remove after debugging
--		self:EnforceProductsCannotHaveCostSetting()

		self:ResetSkillCaches()
		self:ApplyFilters()
		self:UpdateProductOptions()
		self:UpdateResourceCosts()
		self:UpdateStatistics()
		AceConfigRegistry:NotifyChange("Capitalist")
	end
	
	self:UpdateRestockStack(profession)

	self:DebugExit("ScanTradeSkillFrame")
end

function Capitalist:BuildList()
	self:DebugEntry("BuildList")

	self.list = {}

	-- Generally, it's "simple bucketing" -- 1 full product name per bucket
	self.scanBuckets = {}
	-- Optimization 1: do "Glyph of " in batch
	local productsIncludeGlyph = false
	-- Optimization 2: do "Scroll of Enchant " in batch
	local productsIncludeScroll = false

	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			self:DebugMisc(profession)

			for skillName, skillData in pairs(professionData) do
				local productName = self:ProductNameFromSkillName(profession, skillName)
				if (productName and
					self:ProductIsVisible(profession, skillName) and
					self:ProductIsEnabled(profession, skillName)) then

					local applyOptimizationGlyph = false
					local applyOptimizationScroll = false
					if (productName:find(L["Glyph of "])) then
						applyOptimizationGlyph = true
					end
					if (productName:find(L["Scroll of Enchant "])) then
						applyOptimizationScroll = true
					end

					self.list[productName] = { profession = profession, skillName = skillName }

					if (self:ProductIsEnabled(profession, skillName)) then
						if (applyOptimizationGlyph) then
							productsIncludeGlyph = true
						elseif (applyOptimizationScroll) then
							productsIncludeScroll = true
						else
							table.insert(self.scanBuckets, productName)
						end
					end
				end
			end
		end
	end

	if (productsIncludeGlyph) then
		table.insert(self.scanBuckets, 1, L["Glyph of "])
	end
	if (productsIncludeScroll) then
		table.insert(self.scanBuckets, 1, L["Scroll of Enchant "])
	end

	self:DebugExit("BuildList")
end
